/**
 * This module is responsible for bootstrapping the core application resources.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 8 Oct, 2013
 * Updated: 19 May, 2014
 */
var bootstrapper = (function () {
	var guid = (function() {
			  function s4() {
				return Math.floor((1 + Math.random()) * 0x10000)
						   .toString(16)
						   .substring(1);
			  }
			  return function() {
				return s4() + s4() + '-' + s4() + '-' + s4() + '-' +
					   s4() + '-' + s4() + s4() + s4();
			  };
			})();
			
    var userMode = false;
    var STORAGE_KEY = 'bundleVersion';

    // Update Information DIV
    var updateDIV = document.createElement('div');
    updateDIV.classList.add('applicationUpdate');

    /**
     * Used to display error messages based on the visibility flag.
     *
     * @param message The message
     * @private
     */
    function _showError(message) {
        if (userMode) {
            alert(message);
        }
    }

    /**
     * Check internet connection status.
     *
     * @param exitOnFail
     * @private
     */
    function _checkConnection(exitOnFail) {
        var message = "This app requires a connection.\nPlease reconnect and try again.";
        var appExitCallback;

        // Override the param to become a function that exists the app
        if (exitOnFail && navigator && navigator.app) {
            appExitCallback = navigator.app.exitApp;
        }

        if (navigator.connection.type == Connection.NONE) {
            navigator.notification.confirm(message, appExitCallback, "Offline", "OK");
            return false;
        }

        return true;
    }

    /**
     * Perform AJAX GET on the provided URL.
     *
     * @param url The request URL
     * @param onWin Callback on success
     * @param onFail Callback on failure
     * @private
     */
    function _get(url, onWin, onFail) {
        var xmlhttp = new XMLHttpRequest();
        xmlhttp.open("GET", encodeURI(url), false);

        xmlhttp.onload = function () {
            onWin(xmlhttp.responseText);
        };

        xmlhttp.onerror = onFail;

        try {
            xmlhttp.send();
        } catch (err) {
            onFail();
            console.error(err);
        }
    }

    /**
     * Perform AJAX GET on the provided URL to fetch Binary data.
     *
     * @param url Url to request
     * @param onWin Callback on success
     * @param onFail Callback on failure
     * @private
     */
    function _getBinary(url, onWin, onFail) {
        var xmlhttp = new XMLHttpRequest();
        xmlhttp.open("GET", url, true);
        xmlhttp.responseType = "arraybuffer";

        // Download successful
        /*xmlhttp.onload = function () {
            onWin(xmlhttp.response);
        };*/

        xmlhttp.onreadystatechange = function () {
            if (xmlhttp.readyState == 4) {
                //alert("xml http on readystate change" + xmlhttp.readyState);
                onWin(xmlhttp.response);
            }
        };

        // Download failed
        xmlhttp.onerror = onFail;

        try {
            xmlhttp.send();
        } catch (err) {
            onFail();
            console.error(err);
        }
    }

    /**
     * Checks for version difference on the server.
     *
     * @param onDiff Callback when there's a new version
     * @param onComplete When the done testing for version difference regardless of the state
     */
    function _versionCheck(onDiff, onComplete) {
        document.body.appendChild(updateDIV);
        var url = window.host + '/appversion.txt?v=' + new Date();

        _get(url, function (data) {
            try {
                var versions = JSON.parse(data);
                var bundleVersion = localStorage.getItem(STORAGE_KEY);

                if (versions && bundleVersion != versions.version) {
                    updateDIV.innerText = 'A new application update was found';

                    // Download all scripts all over again
                    onDiff(versions.url, function () {
                        // Reset the cache
                        localStorage.setItem(STORAGE_KEY, versions.version);
                    });
                } else {
                    document.body.removeChild(updateDIV);
                    onComplete();
                }
            } catch (err) {
                console.error('Failed to parse received version data', data, err);
                document.body.removeChild(updateDIV);
                _showError("Failed to check for app updates at the moment.\nPlease try again later.");

                // Failed to fetch the application version, proceed to run the application
                // In user mode we will not go to main, since we're already there
                if (!userMode) {
                    _proceed();
                }
            }
        }, function (err) {
            console.error('Failed to check for app updates at the moment', err);
            document.body.removeChild(updateDIV);
            _showError("Failed to check for app updates at the moment.\nPlease try again later.");

            // Failed to fetch the application version, proceed to run the application
            // In user mode we will not go to main, since we're already there
            if (!userMode) {
                _proceed();
            }
        });
    }

    /**
     * Downloads the application's JS resources.
     * This method will first check for version differences
     * then if there's a diff, it will delete the current app
     * folder and then finally download all over again.
     *
     * @param url The url to download the resource from
     * @param win Callback to execute when the update is complete
     */
    function _downloadAndProceed(url, win) {
        updateDIV.innerText = 'Downloading application update ..';

        fsManager.downloadResourceVersion2(url, function (fileEntry) {
            updateDIV.innerText = 'Update was downloaded successfully ..';

            fsManager.deleteAppDir(function () {
                fsManager.getAppCacheRootDir(function (path) {
                    fsManager.unzipResources(fileEntry, path, function () {
                        updateDIV.innerText = 'Your application has been updated successfully ..';

                        if (win) {
                            win();
                        }

                        _proceed();
                    }, function () {
                        console.error('Failed to unzip the new bundle');
                        document.body.removeChild(updateDIV);
                        _showError("Failed to complete update, please try again later");

                        // Unzip failed, hence we'll use the APK's bundle
                        localStorage.removeItem(STORAGE_KEY);
                        _init(true);
                    });
                }, function () {
                    console.error('Failed to fetch the app path so that we unzip the new bundle, re-expanding the bundled zip');
                    document.body.removeChild(updateDIV);
                    _showError("Failed to complete update, please try again later");
                    localStorage.removeItem(STORAGE_KEY);
                    _init(true);
                });
            },
                // If somehow we fail when deleting the dir,
                function () {
                    console.error('Failed to delete the old app path, going to main');
                    document.body.removeChild(updateDIV);
                    _showError("Failed to complete update, please try again later");
                    _proceed();
                });
        }, function () {
            console.log('Failed to download app update, going to main');
            document.error.removeChild(updateDIV);
            _showError("Failed to download update, please try again later");

            // We failed to download the new bundle, proceed to load the application
            _proceed();
        });
    }

    /**
     * Proceed to the main page.
     */
    function _proceed() {
        // Acquire the app root path, then redirect to the next page
        fsManager.getAppCacheRootDir(function (path) {
            localStorage.setItem('fsPath', path.toURL().replace('file://', ''));
            window.location = 'index.html?preview=true&v='+guid();
        });
    }

    /**
     * Checks for a newer version of the bundle on the server
     * if any it will download and un-pack it.
     */
    function _contactServerForBundle() {
        _versionCheck(
            // On version difference
            _downloadAndProceed,
            // On version not changed
            _proceed);
    }

    /**
     * Force update check wth the server.
     *
     * @private
     */
    function _forceUpdate() {
        if (!_checkConnection(false)) {
            return false;
        }

        userMode = true;

        _versionCheck(
            // On version difference
            _downloadAndProceed,
            // On version not changed
            function () {
                alert('You are using the latest version');
            });
    }

    /**
     * Initialize the application.
     *
     * @param bypassNewVersionCheck If true, once the local APK bundle is unzipped, bypass checking for a new version
     */
    function _init(bypassNewVersionCheck) {
        // No connection, exit the app and do not proceed
        if (!_checkConnection(true)) {
            return false;
        }

        // Check if the app path exists, if yes proceed to check
        // for the version check and so on
        fsManager.isAppPathExist(
            function () {
                _initializeAppBundle(bypassNewVersionCheck);
            },
            // Else, extract the bundle
            function () {
                _extractLocalBundle(bypassNewVersionCheck);
            });
    }

    /**
     * Initialize the application.
     *
     * @param bypassNewVersionCheck If true, once the local APK bundle is unzipped, bypass checking for a new version
     */
    function _initializeAppBundle(bypassNewVersionCheck) {
        // If there's no version in the localStorage, then we're running
        // the app for the very first time, in which case we need to
        // fetch the bundled version file and un-pack the included bundle as well
        if (!localStorage.getItem(STORAGE_KEY)) {
            _extractLocalBundle(bypassNewVersionCheck);
        }
            // Check for new version on server
            // otherwise, we'll just launch the app
        else {
            _contactServerForBundle();
        }
    }

    /**
     * Extracts the bundled app zip and checks for a new version
     * based on the provided parameter's value
     * 
     * @param bypassNewVersionCheck If true, once the local APK bundle is unzipped, bypass checking for a new version
     */
    function _extractLocalBundle(bypassNewVersionCheck) {
        // Save the phone.zip from the APK into the local FS
        // then unzip it
        fsManager.getAppCacheRootDir(function (path) {
            _getBinary('phone.zip', function (binData) {
                fsManager.writeFile('phone.zip', binData, function (fileEntry) {
                    fsManager.unzipResources(fileEntry, path, function () {
                        // Add the current bundle version from the APK into the localStorage
                        _get('appversion.txt', function (data) {
                            // Set the bundled version into the locaStorage
                            var versions = JSON.parse(data);
                            localStorage.setItem(STORAGE_KEY, versions.version);

                            // Proceed either to check for a newer version or to
                            // the main page directly
                            _initExitHandler(bypassNewVersionCheck, false);
                        },
                         // Failed to fetch local app version
                         function (err) {
                             _initExitHandler(bypassNewVersionCheck, false, err);
                         });
                    },
                    // Failed to unzip
                    function () {
                        _initExitHandler(bypassNewVersionCheck, true);
                    });
                }, function (error) {
                    _initExitHandler(bypassNewVersionCheck, true, error);
                });
            }, function (error) {
                _initExitHandler(bypassNewVersionCheck, true, error);
            });
        }, function (error) {
            _initExitHandler(bypassNewVersionCheck, true, error);
        });
    }

    /**
     * Util function to handle terminal cases in the init function.
     * 
     * @param bypassNewVersionCheck Should bypass version check
     * @param terminate Should terminate app when the bypassNewVersionCheck param is false
     * @param error Any possible error object
     */
    var _initExitHandler = function (bypassNewVersionCheck, terminate, error) {
        // Failed, contact server for bundle and attempt to unzip it
        if (!bypassNewVersionCheck) {
            _contactServerForBundle();
        } else {
            if (terminate) {
                navigator.notification.confirm("Failed to load app data.\nApplication will exit!", navigator.app.exitApp, "Offline", "OK");
            } else {
                _proceed();
            }
        }

        console.error('Failed to write file', error);
    };

    return {
        initialize: _init,
        contactServerForBundle: _contactServerForBundle,
        forceUpdate: _forceUpdate
    };
})();