/**
 * This module provides file download facility for use with offline resource caching.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 8 Oct, 2013
 * Updated: 19 May, 2014
 */
var fsManager = (function () {
    // Static fields representing all sorts of file types
    var _DIR_SCRIPTS = "js";
    var _DIR_CSS = "css";
    var _DIR_IMAGES = "images";
    var _DIR_HTML = "html";
    var _DIR_TEMPLATES = "templates";

    // Fail safe for Android
    window.requestFileSystem = window.requestFileSystem || window.webkitRequestFileSystem;

    /**
     * Returns the appropriate message for the provided file error code.
     *
     * @param code The error code
     */ 
    var _getFileErrorMessage = function (code) {
        var msg = 'Unknown Error Code';
        
        switch (code) {
            case FileError.QUOTA_EXCEEDED_ERR:
                msg = 'Quota exceeded error';
                break;
            case FileError.NOT_FOUND_ERR:
                msg = 'File not found error';
                break;
            case FileError.SECURITY_ERR:
                msg = 'Security error';
                break;
            case FileError.INVALID_MODIFICATION_ERR:
                msg = 'Invalid modification error';
                break;
            case FileError.INVALID_STATE_ERR:
                msg = 'Invalid state error';
                break;
        }
        
        return msg;
    };
    
    /**
     * Returns the appropriate message for the provided file transfer error code.
     * 
     * @param code The error code
     */
    var _getFileTransferErrorMessage = function (code) {
        var msg = 'Unknown Error Code';
        
        switch (code) {
            case FileTransferError.FILE_NOT_FOUND_ERR:
                msg = 'File not found error';
                break;
            case FileTransferError.INVALID_URL_ERR:
                msg = 'Invalid URL error';
                break;
            case FileTransferError.CONNECTION_ERR:
                msg = 'Connection error';
                break;
            case FileTransferError.ABORT_ERR:
                msg = 'Abort error';
                break;
        }

        return msg;
    };
    
    /**
     * Fetches the app cache dir with the specified app name.
     *
     * @param callback The callback
     * @param fail Callback for failure
     */
    var _getAppCacheRootDir = function (callback, fail) {
        try {
            window.requestFileSystem(window.PERSISTENT, 0, function (fileSystem) {
                // Create the root path
                fileSystem.root.getDirectory('kaleoapp',
                    { create: true, exclusive: false },
                    function (parent) {
                        parent.setMetadata(function () { console.log('set meta data success'); }, function () { console.log('set meta data failed'); }, { "com.apple.MobileBackup": 1 });
                        if (callback) {
                            callback(parent);
                        }
                    },
                    function (error) {
                        console.error("Failed to create app dir: " + error.code);
                    });
            }, function (err) {
                console.error('Failed to request file system.'
                              + '\n> Error Code: ' + err.code + ' (' + _getFileErrorMessage(err.code) + ')'
                              + '\n> Error Code: ' + err.code
                              + '\n> Error Source: ' + err.source
                              + '\n> Error Target: ' + err.target
                              + '\n> Error Status: ' + err.status
                              + '\n> Error Body: ' + err.body
                              + '\nRAW> = ', err);

                if (fail) {
                    fail();
                }
            });
        } catch (err) {
            console.error('Failed to get FS', err);

            if (fail) {
                fail();
            }
        }
    };
    
    /**
     * Checks if the app path exists, and invokes the appropriate callbacks.
     * 
     * @param win Success callback
     * @param fail Failure callback
     */
    var _isAppPathExist = function (win, fail) {
        try {
            if (window && window.requestFileSystem) {
                // Success callback for FS request
                var fsSuccess = function (fileSystem) {
                    //alert("after requestFileSystem _isAppPathExist");
                    //alert("before getDirectory kaleoapp _isAppPathExist");
                    fileSystem.root.getDirectory('kaleoapp', { create: false, exclusive: false }, function (parent) {
                        parent.setMetadata(function () { console.log('set meta data success 1'); }, function () { console.log('set meta data failed 1'); }, { "com.apple.MobileBackup": 1 });
                        win();

                    }
                        , fail);

                };
                
                // Request the FS
                window.requestFileSystem(window.PERSISTENT, 0, fsSuccess, fail);
            } else {
                if (fail) {
                    fail();
                }
            }
        } catch (err) {
            if (fail) {
                fail();
            }
        }
    };

    /**
     * Downloads a resource from the specified URI into
     * the app cache path using the target file name
     *
     * @param uri The URI to download the resource from
     * @param complete when the download is complete
     * @param fail When download fails
     */
    var _downloadResource = function (uri, complete, fail) {
        _getAppCacheRootDir(function (path) {
            var fileTransfer = new FileTransfer();

            var toPath = path.toURL() + '/' + uri.substring(uri.lastIndexOf('/') + 1);

            // On iOS use native path
//            if (device.platform && device.platform.toLowerCase() == 'ios') {
//                toPath = path.toNativeURL() + '/' + uri.substring(uri.lastIndexOf('/') + 1);
//            }

            fileTransfer.download(
                encodeURI(uri),
                toPath,
                function (entry) {
                    if (complete) {
                        complete(entry);
                    }
                }, function (error) {
                    console.error('Failed to download resource.'
                                  + '\n>>> error code = ' + error.code + ' (' + _getFileTransferErrorMessage(error.code) + ')'
                                  + '\n>>> error source = ' + error.source
                                  + '\n>>> error target = ' + error.target
                                  + '\n>>> error status = ' + error.status
                                  + '\n>>> error body = ' + error.body
                                  + '\nRAW> = ', error);

                    if (fail) {
                        fail();
                    }
                },
                true);
        });
    };


    var _downloadResourceVersion2= function (uri, complete, fail) {
        //alert("call download resource1");
        _getBinaryFromServer(uri, function (binData) {
            //alert("get binary from download resource1 success");
            fsManager.writeFile('phone1.zip', binData, function (fileEntry) {
                //alert("write file from download resource1 success");
                if (complete) {
                    complete(fileEntry);
                }
            }, function () { alert("write file from download resource1 failed") });
        }, function () { alert("get binary from download resource1 failed"); });
    };

    var _getBinaryFromServer = function (url, onWin, onFail) {
        //alert('Get Binary');
        var xmlhttp = new XMLHttpRequest();
        xmlhttp.open("GET", url, true);
        xmlhttp.responseType = "arraybuffer";

        // Download successful
        xmlhttp.onload = function () {
            //alert("xmlhttp onload");
        };


        xmlhttp.onreadystatechange = function () {
            if (xmlhttp.readyState == 4) {
                //alert("xml http on readystate change" + xmlhttp.readyState);
                onWin(xmlhttp.response);
            }
        };

        // Download failed
        xmlhttp.onerror = onFail;

        try {
            xmlhttp.send();
        } catch (err) {
            alert("error xmlhttp1" + err);
            onFail();
            console.error(err);
        }
    };

    /**
     * Write ByteArray content to a file on the apps local path.
     *
     * @param fileName The name of the new file to create
     * @param binData The ByteArray data to write
     * @param onWin When the file is written
     * @param onFail When writing fails
     */
    var _writeFile = function (fileName, binData, onWin, onFail) {
         _getAppCacheRootDir(function(path) {
            path.getFile(fileName,
                { create: true, exclusive: false },
                function (fileEntry) {
                    fileEntry.createWriter(function (writer) {
                        // Write success
                        writer.onwrite = function () {
                            onWin(fileEntry);
                        };

                        // Write fail
                        writer.onerror = onFail;

                        // Write it!
                        writer.write(binData);
                    }, function(error) {
                        console.error("Failed to create writer", error);
                    });
                }, function (error) {
                    console.error("Failed to write file", error);
                });
         });
    };

    /**
     * Looks for missing files in a given path.
     *
     * @param filesArray The array of file names to compare with
     * @param win The callback to invoke on success
     * @param fail The callback to invoke on failure
     */
    var _checkForMissingFiles = function (filesArray, win, fail) {
        _getAppCacheRootDir(function (path) {
            var reader = path.createReader();
            reader.readEntries(function (entries) {
                var count = 0;

                for (var i = 0; i < filesArray.length; i++) {
                    for (var j = 0; j < entries.length; j++) {
                        if (filesArray[i] == entries[j].name) {
                            count++;
                        }
                    }
                }

                if (count < filesArray.length) {
                    if (fail) {
                        fail();
                    }
                } else {
                    if (win) {
                        win();
                    }
                }
            }, undefined);
        });
    };

    /**
     * Attempts to delete the app directory and all of its contents
     *
     * @param win The on success callback
     * @param fail The on failure callback
     */
    var _deleteAppDir = function (win, fail) {
        _getAppCacheRootDir(function (path) {
            window.resolveLocalFileSystemURL(path.toURL(),
                function (fileEntry) {
                    if (fileEntry.isDirectory) {
                        fileEntry.getDirectory("phone", {create: false, exclusive:false}, function (parent) {
                            parent.removeRecursively(
                                function () {
                                    if (win) {
                                        win();
                                    }
                                },
                                function () {
                                    console.error('Failed to delete path: ', path);

                                    if (fail) {
                                        fail();
                                    }
                                });
                        }, function () {
                            console.warn("Unzipped phone path not found, nothing to delete!");
                            
                            if (win) {
                                win();
                            }
                        });
                    }
                },
                // Path does not exist, call the win
                function () {
                    console.error('Path does not exist, proceed with win', path);
                    
                    if (win) {
                        win();
                    }
                });
        });
    };

    /**
     * Unzips a compressed archive into storage.
     *
     * @param fromPath The archive file
     * @param toPath The path to extract to
     * @param win
     * @param fail
     */
    var _unzipResources = function (fromFile, toPath, win, fail) {
        // XXX Hack to overcome an issue with the zip plugin version 1.0.0
        var zipFile = fromFile.toURL().replace('file://', '');
        var destPath = toPath.toURL().replace('file://', '');
        
        try {
            zip.unzip(zipFile, destPath, function (status) {
                if (status == -1) {
                    console.error('Failed to unzip file (from, to)', zipFile, destPath);
                    
                    if (fail) {
                        fail();
                    }
                } else {
                    if (win) {
                        win();
                    }
                }
            });
        } catch (err) {
            console.error('Failed to unzip', err);
        }
    };

    return {
        downloadResource: _downloadResource,
        writeFile: _writeFile,
        checkForMissingFiles: _checkForMissingFiles,
        deleteAppDir: _deleteAppDir,
        getAppCacheRootDir: _getAppCacheRootDir,
        isAppPathExist: _isAppPathExist,
        unzipResources: _unzipResources,
        downloadResourceVersion2 : _downloadResourceVersion2,
        DIR_SCRIPTS: _DIR_SCRIPTS,
        DIR_CSS: _DIR_CSS,
        DIR_IMAGES: _DIR_IMAGES,
        DIR_HTML: _DIR_HTML,
        DIR_TEMPLATES: _DIR_TEMPLATES
    };
})();