﻿/**
 * This module is responsible for template loading into the main view.
 *
 * The single page contains a div with the id 'content', and this module
 * will load what ever template under that div.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 28 Aug, 2013
 * Updated: 24 Mar, 2014
 */
define('TemplateLoader',
['jquery', 'LocalModel', 'TemplatesHelper'],
function ($, model, thelper) {
    // Reset the background of the body
    // this will be called only once, when this file loads
    document.body.style.background = '';
    //document.getElementById('loading').style.display = 'none';

    var TEMPLATES_CACHE = {};

    /**
     * Loads the widget's page, returns a promise.
     *
     * @param type The type of template we're dealing with
     */
    var _loadPage = function (type) {
        var promise = $.Deferred();
        var pagePromise = thelper.getResourceURI('page', type);

        pagePromise.done(function (resource) {
            thelper.getResourceInText(resource, function (data) {
                if (data && data.trim().length > 0) {
                    data = data.replace(/_var_url/gi, window.host);

                    // To avoid undefined strings in the cache
                    TEMPLATES_CACHE[type] = TEMPLATES_CACHE[type] ? TEMPLATES_CACHE[type] += data : data;
                }
                
                promise.resolve();
            });
        }).always(function () {
            promise.resolve();
        });

        return promise;
    };

    /**
     * Loads the widget's scripts, returns a promise.
     *
     * @param template The jQuery template node to load the script into
     * @param type The type of template we're dealing with
     * @param callback A callback to invoke after the script has been loaded
     */
    var _loadScripts = function (template, type, callback) {
        var jsPromise = thelper.getResourceURI('js', type);

        jsPromise.done(function (jsResource) {
            if (jsResource !== undefined && jsResource.length > 0) {
                // Reverse the resource array
                jsResource.reverse();

                // Define a function to load the scripts in order
                // since the scripts are being added dynamically
                // modern browsers will load them asynchronously
                // this function guarantees load order
                var loader = function () {
                    // Are we done yet?
                    if (jsResource.length == 0) {
                        if (callback) {
                            callback();
                        }

                        return true;
                    }

                    // Get the next script
                    var script = jsResource.pop();

                    // Load it and when done, call self
                    thelper.addScript(script, template, loader);
                };

                // Start loading the scripts
                loader();
            }
            // Else, there were no JS files, invoke the callback if any
            else if (callback) {
                callback();
            }
        });
    };

    /**
     * Loads the widget's styles, returns a promise.
     *
     * @param type The type of template we're dealing with
     */
    var _loadCSS = function (type) {
        var promise = $.Deferred();
        var cssPromise = thelper.getResourceURI('css', type);
        
        cssPromise.done(function (cssResource) {
            if (cssResource) {
                for (var i = 0; i < cssResource.length; i++) {
                    thelper.getResourceInText(cssResource[i], function (data) {
                        if (data && data.trim().length > 0) {
                            data = data.replace(/_var_url/gi, window.host);
                            data = '<style>' + data + '</style>';

                            // To avoid undefined strings in the cache
                            TEMPLATES_CACHE[type] = TEMPLATES_CACHE[type] ? TEMPLATES_CACHE[type] += data : data;
                        }
                    });
                }
            }
        }).always(function () {
            // We can safely call this here, because the AJAX
            // call has been set to be synchronous
            promise.resolve();
        });

        return promise;
    };

    /**
     * Loads the widget's fragments, returns a promise.
     *
     * @param type The type of template we're dealing with
     */
    var _loadFragments = function (type) {
        var promise = $.Deferred();
        var fragmentPromise = thelper.getResourceURI('fragments', type);
        
        fragmentPromise.done(function (fragment) {
            if (fragment !== undefined) {
                for (var i = 0; i < fragment.length; i++) {
                    thelper.getResourceInText(fragment[i], function (data) {
                        if (data) {
                            // To avoid undefined strings in the cache
                            TEMPLATES_CACHE[type] = TEMPLATES_CACHE[type] ? TEMPLATES_CACHE[type] += data : data;
                        }
                    });
                }
            }
        }).always(function () {
            // We can safely call this here, because the AJAX
            // call has been set to be synchronous
            promise.resolve();
        });

        return promise;
    };

    /**
     * Adds a new template into the view-port, based on the type.
     * 
     * @param $template The jQuery object for the template DOM object
     * @param $content The jQuery object for the template container
     * @param type The type of template to load
     * @param callback A callback to execute once loading is complete
     */
    var _addWidgetTemplate = function ($template, $content, type, callback) {
        // Remove the current template
        $template.remove();

        // Append the new one based on the passed type
        $content.append(TEMPLATES_CACHE[type] + '</div>');


        // Re-select the template div since the one that
        // is currently selected belongs to the previous one
        $template = $("#template");

        // Show the template
        $template[0].classList.add('fadeIn');

        // Apply the proper background style based on the loaded
        // template, special handling for home
        if (type === 'home') {
            $template[0].classList.add('homeBG');
        } else if (type !== 'text') {
            $template[0].classList.add('widgetBG');
        }

        // Apply the icon styles
        setTimeout(thelper.applyIconStyles, 0);

        // Load scripts for page
        _loadScripts('template', type, function () {
            if (callback) {
                setTimeout(callback, 0);
            }
        });
    };

    /**
     * Loads a template from the pre-defined template files
     *
     * @param type The widget type
     * @param callback A callback function to call once load is complete
     */
    var _loadTemplate = function (type, callback) {
        var deferred = $.Deferred();
        var $content = $("#content");
        var $template = $("#template");

        // If the template is not already loaded, do load
        if ($content.html().replace(/\s+/g, '').length <= 0 || $template.attr('data-template') != type) {
            // Template not cached before, cache it the add it
            if (!TEMPLATES_CACHE[type]) {
                // Add the template container, we need this so that we can properly
                // handle knockout bindings on a level that can be easily removed
                //
                // Add an extra template type flag so that we know
                // that we need to switch templates when the user chooses another one
                TEMPLATES_CACHE[type] = '<div id="template" data-template="' + type + '" class="template" style="opacity: 0;">';

                // Guarantee CSS loading before proceeding
                var cssPromise = _loadCSS(type);
                cssPromise.done(function () {
                    // Guarantee Fragment loading before proceeding
                    var fragmentsPromise = _loadFragments(type);
                    fragmentsPromise.done(function () {
                        // Guarantee page loading before proceeding
                        var pagePromise = _loadPage(type);
                        pagePromise.done(function (pageData) {
                            _addWidgetTemplate($template, $content, type, callback);
                        });
                    });
                });
            }
            // Use cached version
            else {
                _addWidgetTemplate($template, $content, type, callback);
            }
        }
        // Execute the call back responsible for binding
        else {
            if (callback) {
                callback();
            }
        }

        return deferred;
    };

    /**
     * Add the accessory template to the container.
     * 
     * @param $template The template jQuery object
     * @param templateName The name of the template to load
     * @param target The target div id to add the template to
     * @param callback A callback to execute when the template is loaded
     */
    var _addAccessoryTemplate = function ($template, templateName, target, callback) {
        // Hide and Remove the current screen
        $template.empty();

        // Append the new one based on the passed type
        $template.append(TEMPLATES_CACHE[templateName]);
        setTimeout(thelper.applyIconStyles, 0);

        _loadScripts(target, templateName, function () {
            if (callback) {
                setTimeout(callback, 0);
            }
        });
    };

    /**
     * Load an accessory component template, such as the Menu
     *
     * @param templateName The name of the template to load
     * @param target The Id of the div to add the template to
     * @param callback A callback to invoke after loading is done
     */
    var _loadAccessory = function (templateName, target, callback) {
        var deferred = $.Deferred();
        var $template = $("#" + target);

        if (!TEMPLATES_CACHE[templateName]) {
            _loadCSS(templateName);

            var pagePromise = _loadPage(templateName);

            pagePromise.done(function (pageData) {
                _addAccessoryTemplate($template, templateName, target, callback);
            });
        } else {
            _addAccessoryTemplate($template, templateName, target, callback);
        }

        return deferred;
    };

    // Expose
    return {
        loadTemplate: _loadTemplate,
        loadAccessory: _loadAccessory
    }
});