/**
 * This module is responsible for loading widgets into view.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 28 Aug, 2013
 * Updated: 4 Sep, 2013
 */
define('WidgetManager',
['TemplateLoader', 'DataFetcher', 'knockout', 'komapping', 'jquery', 'TemplatesHelper', 'LocalModel'],
function (loader, data, ko, kom, $, thelper, model) {
    // Map kom to ko.mapping, if not mapped already
    if (ko.mapping === undefined) {
        ko.mapping = kom;
    }

    // XXX: Used for benchmarking
    var startTimeWidget;
    var startTimeAccessories;

    // The current widget slide index on the home page
    var currentSlideIndex = 0;

    // History stack
    var history = [];

    // Ajax progress spinner configuration
    var spinner = document.getElementById('spinner');

    $(document).ajaxStart(function () {
        spinner.style.display = 'block';
    }).ajaxStop(function () {
        spinner.style.display = 'none';
    });

    /**
     * Utility function to check if a binding to an element is already in place.
     */
    var isBound = function (div) {
        if (div) {
            return !!ko.dataFor(div);
        }

        // Say it is bound in case the template is not found to avoid invalid bindings
        return true;
    }

    /**
     * Pushes a new item to the history stack.
     *
     * @param item An object of the form: { 'id': <value>, 'type': <type> }
     */
    var _historyPush = function (item) {
        // Don't push duplicates to the top
        if (_historyTop() != item) {
            history.push(item);
        }
    }

    /**
     * Pops the top item from the history stack.
     */
    var _historyPop = function () {
        return history.pop();
    }

    /**
     * Returns the top of the history stack.
     */
    var _historyTop = function () {
        return history[history.length - 1];
    }

    /**
     * Returns the expected previous item in the history without popping it.
     */
    var _historyPreviousItem = function () {
        if (history.length >= 2) {
            return history[history.length - 2];
        }
    }

    /**
     * Clears all items in history.
     */
    var _historyClear = function () {
        return history.length = 1;
    }

    /**
     * Returns the size of the history.
     */
    var _historySize = function () {
        return history.length;
    }

    /**
     * Returns true if the history is empty.
     */
    var _historyEmpty = function () {
        if (history.length <= 1) {
            return true;
        }

        return false;
    }

    /**
     * Navigate back in history.
     */
    var _navigateBack = function () {
        // If the history stack has more than one item
        // pop the top which would normally be the current page
        // keep in mind that we'll always have home as the first
        // item in history
        if (history.length > 1) {
            // Remove the current page from the history
            _historyPop();
        }

        // Navigate to the item on the top now
        var item = _historyTop();

        // If there is an item, navigate back
        if (item) {
            _load(item.id, item.type);
        }
    }

    /**
     * Load the widget into the main view.
     *
     * @param widgetId The id for the widget instance on the API server
     * @param widgetType The type of widget to load
     * @param callback An optional function to calls once the load operation
     *        is complete but just before ko bindings are applied
     */
    var _load = function (widgetId, widgetType, callback) {
        if (ENV === 'dev') {
            startTimeWidget = new Date().getTime();
        }

        if (navigator.connection.type == Connection.NONE) {
            navigator.notification.alert("You're device is offline, please connect", null, "Offline", "OK");
        } else {
            // Fetch the data from the API server
            data.get(widgetId, widgetType, function (data) {
                // load the template and apply ko bindings
                _loadWithModel(data, model.get(widgetType), widgetType, callback);
            });
        }
    }

    /**
     * Loads the menu template in case the passed widgetType is home.
     *
     * @param widgetType The type of widget to load
     */
    var _loadAccessories = function (widgetType, data, modelObject) {
        if (ENV === 'dev') {
            startTimeAccessories = new Date().getTime();
        }

        // Load the media player template only if it wasn't loaded
        if (document.getElementById('mediaPlayer').childNodes.length <= 0) {
            loader.loadAccessory(thelper.TEMPLATE_KEYS.mediaPlayer, 'mediaPlayer', function () {
                if (ENV === 'dev') {
                    console.debug("Media player loading complete @ %d MS", new Date().getTime() - startTimeAccessories);
                }
            });
        }

        if (ENV === 'dev') {
            startTimeAccessories = new Date().getTime();
        }

        // Load the menu if the widgetType is home
        if (widgetType == thelper.TEMPLATE_KEYS.home) {
            // Load the menu template
            loader.loadAccessory(thelper.TEMPLATE_KEYS.menu, 'menu', function () {
                // Apply the binding if not already applied
                // It will be already applied while in design mode
                // where editing a widget will result in a sequence
                // of postMessages to update the UI
                var templateDiv = document.getElementById('menuButtons');

                if (!isBound(templateDiv)) {
                    ko.applyBindings(modelObject, templateDiv);
                }

                ko.mapping.fromJSON(data, {}, modelObject);

                if (ENV === 'dev') {
                    console.debug("Menu loading complete @ %d MS", new Date().getTime() - startTimeAccessories);
                }
            });
        }
    }

    /**
     * Loads a template while receiving the data and model to bind to as parameters.
     *
     * @param data The data to use for the binding
     * @param modelObject The model to bind to
     * @param widgetType The type of widget to load
     * @param callback An optional function to calls once the load operation
     *        is complete but just before ko bindings are applied
     */
    var _loadWithModel = function (data, modelObject, widgetType, callback) {
        if (!startTimeWidget && ENV === 'dev') {
            startTimeWidget = new Date().getTime();
        }

        if (modelObject == null) {
            console.error("Model cannot be null");
        } else if (widgetType == null) {
            console.error("WidgetType cannot be null");
        } else {
            // Add to history, unless we're calling this while
            // navigating back
            var top = _historyTop();
            var widgetId = 0;

            // Home widget has a special treatment, its Id is always 0
            if (widgetType !== thelper.TEMPLATE_KEYS.home) {
                widgetId = JSON.parse(data).Id;
            }

            // Since this method will be used when invoking
            // history navigate back, we need to be sure
            // we don't push backwards navigation to history
            if (!top || top.id != widgetId) {
                // Push opened widget information to history
                _historyPush({ id: widgetId, type: widgetType });
            }

            // Load the accessories
            _loadAccessories(widgetType, data, modelObject);

            // Load the template
            loader.loadTemplate(widgetType, function () {
                // Apply the binding if not already applied
                // It will be already applied while in design mode
                // where editing a widget will result in a sequence
                // of postMessages to update the UI
                var templateDiv = document.getElementById('template');

                if (!isBound(templateDiv)) {
                    ko.applyBindings(modelObject, templateDiv);
                }

                ko.mapping.fromJSON(data, {}, modelObject);

                // Call initialize on the widget template if available
                if (widgetTemplate && widgetTemplate.init) {
                    widgetTemplate.init();
                }

                // Invoke the callback if defined
                if (callback) {
                    callback();
                }

                if (ENV === 'dev') {
                    console.debug("Widget %s with id %d loading complete @ %d MS", widgetType, widgetId, new Date().getTime() - startTimeAccessories);
                    startTimeWidget = undefined;
                }

                try {
                    if (eventsCallback && (eventsCallback['LoadWidgetWithModel']) && (eventsCallback['LoadWidgetWithModel'].callbacks)) {
                        for (var i = 0; i < eventsCallback['LoadWidgetWithModel'].callbacks.length; i++) {
                            eventsCallback['LoadWidgetWithModel'].callbacks[i](widgetTemplate);
                        }
                    }
                }
                catch (ex) { }
            });
        }
    }

    /**
     * Loads a template while receiving the data and model to bind to as parameters.
     *
     * @param widgetType The type of widget to load
     * @param callback An optional function to calls once the load operation
     *        is complete but just before ko bindings are applied
     */
    var _loadWithoutModel = function (widgetType, callback) {
        if (widgetType == null) {
            console.error("WidgetType cannot be null");
        } else {
            loader.loadTemplate(widgetType, function () {
                // Call initialize on the widget template if available
                if (widgetTemplate && widgetTemplate.init) {
                    widgetTemplate.init();
                }

                // Invoke the callback if defined
                if (callback !== undefined) {
                    callback();
                }

                // Show the new content after the data had been bound
                document.getElementById('template').classList.add('fadeIn');
            });
        }
    }

    /**
     * Sets the current page index of the home page.
     *
     * @param index The index of the current page
     */
    var _setCurrentWidgetSlide = function (index) {
        currentSlideIndex = index;
    }

    /**
     * Go to the current slide based on the cached current slide index value.
     *
     * @param carouselId The Id of the div hosting tha carousel
     */
    var _gotoCurrentWidgetSlide = function (carouselId) {
        var owl = $(carouselId).data('owlCarousel');
        var totalItems = $('.item').length;

        if (currentSlideIndex && currentSlideIndex >= 0) {
            owl.jumpTo(currentSlideIndex);
        } else if (totalItems > 0) {
            // We need to force the jumping to the first slide anyway
            // so that the active style gets applied to the first page
            // failing to do so will only enable the feature to work after
            // the first navigation back and forth
            owl.jumpTo(0);
        }
    }

    var eventsCallback = {};
    var _subscribe = function (eventName, callback) {
        if (!eventsCallback) eventsCallback = {};
        if (!eventsCallback[eventName]) eventsCallback[eventName] = { callbacks: [] };
        eventsCallback[eventName].callbacks.push(callback);
    }

    // Expose
    return {
        show: _load,
        showWithModel: _loadWithModel,
        showWithoutModel: _loadWithoutModel,
        navigateBack: _navigateBack,
        historyPreviousItem: _historyPreviousItem,
        historyEmpty: _historyEmpty,
        historySize: _historySize,
        historyClear: _historyClear,
        setCurrentWidgetSlide: _setCurrentWidgetSlide,
        gotoCurrentWidgetSlide: _gotoCurrentWidgetSlide,
        subscribe: _subscribe
    }
});