/**
 * This module provides file download facility for use with offline resource caching.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 8 Oct, 2013
 * Updated: 11 Oct, 2013
 */
var fsManager = (function () {
    "use strict";

    // Static fields representing all sorts of file types
    var _DIR_SCRIPTS = "templates";
    var _DIR_CSS = "templates";
    var _DIR_IMAGES = "templates";
    var _DIR_HTML = "templates";
    var _DIR_TEMPLATES = "templates";

    /**
     * Fetches the app cache dir with the specified app name
     *
     * @param dirName The name of the app
     * @param callback The callback
     */
    function _getAppCacheRootDir(callback) {
        window.requestFileSystem(LocalFileSystem.PERSISTENT, 0, function (fileSystem) {
            var entry = fileSystem.root;

            // Create the root path
            entry.getDirectory(window.appName,
                { create: true, exclusive: false },
                function (parent) {
                    if (callback != undefined) {
                        callback(parent);
                    }
                },
                function (error) {
                    console.error("Failed to create app dir: " + error.code);
                });
        }, null);
    }

    /**
     * Fetches the resource specific cache dir
     *
     * @param type Type of the resource we are dealing with
     * @param callback The callback
     */
    function _getAppResourceCacheDir(type, callback) {
        _getAppCacheRootDir(function (path) {
            path.getDirectory(type,
                { create: true, exclusive: false },
                function (parent) {
                    callback(parent);
                },
                function (error) {
                    console.error("Failed to create app dir: " + error.code);
                });
        });
    }

    /**
     * Downloads a resource from the specified URI into
     * the app cache path using the target file name
     *
     * @param uri The URI to download the resource from
     * @param targetName The target file name
     */
    function _downloadResource(uri, type, targetName, complete, fail) {
        _getAppResourceCacheDir(type, function (path) {
            var fileTransfer = new FileTransfer();

            console.log('Downloading\nsource = ' + encodeURI(uri) + '\nto ' + path.fullPath + '/' + targetName + '\n');

            fileTransfer.download(
                encodeURI(uri),
                path.fullPath + '/' + targetName,
                function (entry) {
                    console.log('Resource downloaded -> ' + entry.fullPath);

                    if (complete) {
                        complete();
                    }
                },
                function (error) {
                    console.error('Failed to download resource -> ' + targetName
                        + ', \nerror code = ' + error.code
                        + ', \nerror source = ' + error.source
                        + ', \nerror target = ' + error.target
                        + ', \nerror status = ' + error.status
                        + ', \nerror body = ' + error.body + '\n');

                        if (fail) {
                            fail();
                        }
                },
                false
            );
        });
    }

    /**
     * Checks if a given file exists
     *
     * @param fileName The name of the file
     */
    function _isFileExists(type, fileName, win, fail) {
        _getAppResourceCacheDir(type, function (path) {
            window.resolveLocalFileSystemURI(path.fullPath + '/' + fileName,
                function (fileEntry) {
                    if (win) {
                        win();
                    }
                },
                function (evt) {
                    if (fail) {
                        fail();
                    }
                });
        });
    }
    
    /**
     * Looks for missing files in a given path
     *
     * @param type Resource type
     * @param filesArray The array of file names to compare with
     * @param win The callback to invoke on success
     * @param fail The callback to invoke on failure
     */
    function _checkForMissingFiles(type, filesArray, win, fail) {
        _getAppResourceCacheDir(type, function (path) {
            var reader = path.createReader();
            reader.readEntries(function (entries) {
                var count = 0;
                
                for (var i = 0; i < filesArray.length; i++) {
                    for (var j = 0; j < entries.length; j++) {
                        if (filesArray[i] == entries[j].name) {
                            count++;
                        }
                    }
                }
                
                if (count < filesArray.length) {
                    if (fail) {
                        fail();
                    }
                } else {
                    if (win) {
                        win();
                    }
                }
            }, undefined);
        });
    }

    /**
     * Attempts to delete the app directory and all of its contents
     *
     * @param win The on success callback
     * @param fail The on failure callback
     */
    function _deleteAppDir(win, fail) {
        _getAppCacheRootDir(function (path) {
            window.resolveLocalFileSystemURI(path.fullPath,
                function (fileEntry) {
                    fileEntry.removeRecursively(
                        function () {
                            console.log('App path was deleted successfully!');

                            if (win) {
                                win();
                            }
                        },
                        function () {
                            console.error('Failed to delete path: ' + path.fullPath);

                            if (fail) {
                                fail();
                            }
                        });
                },
                // Path does not exist, call the win
                function (evt) {
                    if (win) {
                        win();
                    }
                });
        });
    }

    return {
        downloadResource: _downloadResource,
        isFileExists: _isFileExists,
        checkForMissingFiles: _checkForMissingFiles,
        deleteAppDir: _deleteAppDir,
        getAppCacheRootDir: _getAppCacheRootDir,
        DIR_SCRIPTS: _DIR_SCRIPTS,
        DIR_CSS: _DIR_CSS,
        DIR_IMAGES: _DIR_IMAGES,
        DIR_HTML: _DIR_HTML,
        DIR_TEMPLATES: _DIR_TEMPLATES
    }
})();

if (typeof define === "function" && define.amd) {
    define("fsManager", ['AppConfig'], function (cfg) { return fsManager; });
}