/**
 * This library aims at facilitating communication between windows across domains.
 *
 * It wraps up the sendMessage function while enabling recipients to register different
 * message handlers for the different messages.
 *
 * A message is to be identified by a specific identifier representing its purpose.
 *
 * Author: Waleed Madanat
 * Version: 1.0
 * Created: 16 July, 2013
 * Updated: 21 July, 2013
 */
var kcomm = (function (utility) {
    // An array to hold registered message handlers
    var messageHandlers = [];

    // Indicates the current status, listening or not
    var listening = false;

    /**
     * Handles the onMessage event, it will parse the JSON message and
     * invoke the appropriate MessageHandler sub-class.
     *
     * @param event The message event object
     */
    function onMessageHandler(event) {
        // Make sure the event is not null and that it is a JSON string
        if (event != null && utility.isJSON(event.data)) {
            // Parse the event data into JSON
            var jsonObject = JSON.parse(event.data);

            // Loop over the registered handlers and check if the current
            // can in fact handle the received message, in which case, 
            // call the handle method passing it the message
            for (var i = 0; i < messageHandlers.length; i++) {
                if ("canHandle" in messageHandlers[i]) {
                    if (messageHandlers[i].canHandle(jsonObject)) {
                        messageHandlers[i].handle(jsonObject);
                    }
                }
            }
        } else {
            console.error("Un-supported data type, should be a JSON string");
        }
    }

    return {
        /**
         * Send a message to the target window.
         * Currently this is open for all domains! No restrictions, could be risky!
         *
         * @param msg The message String to send
         * @param target The target window object
         * @returns {string}
         */
        sendMessage: function (msg, target) {
            if (utility.isWindow(target)) {
                target.postMessage(msg, "*");
            } else {
                return "target must be a window!";
            }

        },

        /**
         * Registers a message listener on the current window.
         */
        listen: function () {
            if (!listening) {
                window.addEventListener('message', onMessageHandler, true);

                // Toggle status to indicate that we are now listening
                listening = true;
            }
        },

        /**
         * De-registers the message event listener on the current window.
         */
        stopListening: function () {
            // Remove the event listener
            window.removeEventListener('message', onMessageHandler, true);

            // Toggle status to indicate that we've stopped listening
            listening = false;
        },

        /**
         * Returns the current status, listening or not.
         */
        isListening: function () {
            return listening;
        },

        /**
         * Register a message handler based on a type.
         *
         * @param handlerFn The message handler object
         */
        registerHandler: function(handlerFn) {
            messageHandlers[messageHandlers.length] = handlerFn;
        },

        /**
         * De-register a previously registered message handler.
         *
         * @param handlerFn The message handler object
         */
        removeHandler: function(handlerFn) {
            var arrayLength = messageHandlers.length;

            // Performing a sequential search on the array
            // justifiable since it shouldn't be big
            for (var i = 0; i < arrayLength; i++) {
                if (messageHandlers[i] === handlerFn) {
                    messageHandlers.splice(i, 1);
                }
            }
        },

        MessageHandler: {
            canHandle: function () {
                Console.error("You should not instantiate or call any method from the MessageHandler object");
                throw new Error("Not implemented");
            },

            handle: function () {
                Console.error("You should not instantiate or call any method from the MessageHandler object");
                throw new Error("Not implemented");
            }
        }
    };
})(utility);

if (typeof define === "function" && define.amd) {
    define("kcomm", [], function () { return kcomm; });
}